# aclnnAddcmul&aclnnInplaceAddcmul

## 支持的产品型号
- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型
- aclnnAddcmul和aclnnInplaceAddcmul实现相同的功能，使用区别如下，请根据自身实际场景选择合适的算子。
  - aclnnAddcmul：需新建一个输出张量对象存储计算结果。
  - aclnnInplaceAddcmul：无需新建输出张量对象，直接在输入张量的内存中存储计算结果。
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用 “aclnnAddcmulGetWorkspaceSize” 或者 “aclnnInplaceAddcmulGetWorkspaceSize” 接口获取入参并根据计算流程计算所需workspace大小，再调用 “aclnnAddcmul” 或者 “aclnnInplaceAddcmul” 接口执行计算。
  - `aclnnStatus aclnnAddcmulGetWorkspaceSize(const aclTensor *self, const aclTensor *tensor1, const aclTensor *tensor2,  const aclScalar *value, const aclTensor *out, uint64_t *workspaceSize, aclOpExecutor** executor)`
  - `aclnnStatus aclnnAddcmul(void* workspace, int64 workspaceSize, aclOpExecutor** executor, aclrtStream stream)`
  - `aclnnStatus aclnnInplaceAddcmulGetWorkspaceSize(const aclTensor *selfRef, const aclTensor *tensor1, const aclTensor *tensor2,  const aclScalar *value, uint64_t *workspaceSize, aclOpExecutor** executor)`
  - `aclnnStatus aclnnInplaceAddcmul(void* workspace, int64 workspaceSize, aclOpExecutor** executor, aclrtStream stream)`

## 功能描述

- 算子功能：执行 tensor1 与 tensor2 的逐元素乘法，将结果乘以标量值value并将其添加到输入self。
- 计算公式：

$$
output​=self+ value \times tensor1 \times tensor2
$$

## aclnnAddcmulGetWorkspaceSize

- **参数说明：**

  - self：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与tensor1、tensor2相乘的结果满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与tensor1、tensor2满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与tensor1、tensor2一致。
  - tensor1：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与self、tensor2满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与self、tensor2满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与self、tensor2一致。
  - tensor2：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与self、tensor1满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与self、tensor1满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与self、tensor1一致。
  - value：Host侧的aclScalar，数据类型需要可转换成self与tensor1、tensor2推导后的数据类型。
  - out ：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8，且数据类型需要是self与tensor1、tensor2推导之后可转换的数据类型（[参考说明](common/互转换关系.md)），shape需要是self与tensor1、tensor2 broadcast之后的shape，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与self一致。
  - workspaceSize：返回用户需要在Device侧申请的workspace大小。
  - executor：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```
第一段接口完成入参校验，出现如下场景时报错：
ACLNN_ERR_PARAM_NULLPTR：1. 传入的self、tensor1、tensor2、value或out是空指针。
ACLNN_ERR_PARAM_INVALID：1. self和tensor1、tensor2的数据类型和数据格式不在支持的范围之内。
                         2. self和tensor1、tensor2无法做数据类型推导。
                         3. 推导出的数据类型无法转换为指定输出out的类型。
                         4. self和tensor1、tensor2的shape无法做broadcast。
```

## aclnnAddcmul

- **参数说明：**

  - workspace：在Device侧申请的workspace内存起址。
  - workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnAddcmulGetWorkspaceSize获取。
  - stream：指定执行任务的 AscendCL Stream流。
  - executor：op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceAddcmulGetWorkspaceSize

- **参数说明：**

  - selfRef：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与tensor1、tensor2相乘的结果满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与tensor1、tensor2满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与tensor1、tensor2一致
  - tensor1：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与selfRef、tensor2满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与selfRef、tensor2满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与selfRef、tensor2一致
  - tensor2：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE、BFLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)、INT32、INT64、INT8、UINT8、BOOL，且数据类型需要与selfRef、tensor1满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与selfRef、tensor1满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且[数据格式](common/数据格式.md)需要与selfRef、tensor1一致。
  - value：Host侧的aclScalar，数据类型需要可转换成selfRef与tensor1、tensor2推导后的数据类型。
  - workspaceSize：返回用户需要在Device侧申请的workspace大小。
  - executor：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```
第一段接口完成入参校验，出现如下场景时报错：
ACLNN_ERR_PARAM_NULLPTR：1. 传入的selfRef、tensor1、tensor2或value是空指针。
ACLNN_ERR_PARAM_INVALID：1. selfRef和tensor1、tensor2的数据类型和数据格式不在支持的范围之内。
                         2. selfRef和tensor1、tensor2无法做数据类型推导。
                         3. 推导出的数据类型无法转换为selfRef的类型。
                         4. selfRef和tensor1、tensor2的shape无法做broadcast。
```

## aclnnInplaceAddcmul

- **参数说明：**

  - workspace：在Device侧申请的workspace内存起址。
  - workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnInplaceAddcmulGetWorkspaceSize获取。
  - stream：指定执行任务的 AscendCL Stream流。
  - executor：op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
无

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_addcmul.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);
  // 调用aclrtMemcpy将host侧数据拷贝到device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> selfShape = {4, 2};
  std::vector<int64_t> tensor1Shape = {4, 2};
  std::vector<int64_t> tensor2Shape = {4, 2};
  std::vector<int64_t> outShape = {4, 2};
  void* selfDeviceAddr = nullptr;
  void* tensor1DeviceAddr = nullptr;
  void* tensor2DeviceAddr = nullptr;
  void* outDeviceAddr = nullptr;
  aclTensor* self = nullptr;
  aclTensor* tensor1 = nullptr;
  aclTensor* tensor2 = nullptr;
  aclScalar* value = nullptr;
  aclTensor* out = nullptr;

  std::vector<float> selfHostData = {0, 1, 2, 3, 4, 5, 6, 7};
  std::vector<float> tensor1HostData = {2, 2, 2, 2, 2, 2, 2, 2};
  std::vector<float> tensor2HostData = {2, 2, 2, 2, 2, 2, 2, 2};
  std::vector<float> outHostData = {0, 0, 0, 0, 0, 0, 0, 0};
  float scalarValue = 1.2f;

  // 创建self aclTensor
  ret = CreateAclTensor(selfHostData, selfShape, &selfDeviceAddr, aclDataType::ACL_FLOAT, &self);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建tensor1 aclTensor
  ret = CreateAclTensor(tensor1HostData, tensor1Shape, &tensor1DeviceAddr, aclDataType::ACL_FLOAT, &tensor1);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建tensor2 aclTensor
  ret = CreateAclTensor(tensor2HostData, tensor2Shape, &tensor2DeviceAddr, aclDataType::ACL_FLOAT, &tensor2);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建value aclScalar
  value = aclCreateScalar(&scalarValue, aclDataType::ACL_FLOAT);
  // 创建out aclTensor
  ret = CreateAclTensor(outHostData, outShape, &outDeviceAddr, aclDataType::ACL_FLOAT, &out);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  // 3. 调用CANN算子库API，需要修改为具体的Api名称
  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;
  // 调用aclnnAddcmul第一段接口
  ret = aclnnAddcmulGetWorkspaceSize(self, tensor1, tensor2, value, out, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddcmulGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnAddcmul第二段接口
  ret = aclnnAddcmul(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnAddcmul failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将device侧内存上的结果拷贝至host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(outShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]),
                    outDeviceAddr, size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy resultData from device to host failed. ERROR: %d\n", ret);
            return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("resultData[%ld] is: %f\n", i, resultData[i]);
  }

  // aclnnInplaceAddcmul
  // step3 调用CANN算子库API
  LOG_PRINT("\ntest aclnnInplaceAddcmul\n");
  // 调用aclnnInplaceAddcmul第一段接口
  ret = aclnnInplaceAddcmulGetWorkspaceSize(self, tensor1, tensor2, value, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnInplaceAddcmulGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnInplaceAddcmul第二段接口
  ret = aclnnInplaceAddcmul(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnInplaceAddcmul failed. ERROR: %d\n", ret); return ret);

  // step4（固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // step5 获取输出的值，将device侧内存上的结果拷贝至host侧
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), selfDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(self);
  aclDestroyTensor(tensor1);
  aclDestroyTensor(tensor2);
  aclDestroyTensor(out);
  aclDestroyScalar(value);

  // 7.释放device资源，需要根据具体API的接口定义修改
  aclrtFree(selfDeviceAddr);
  aclrtFree(tensor1DeviceAddr);
  aclrtFree(tensor2DeviceAddr);
  aclrtFree(outDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtFree(stream);
  aclrtFree(context);
  aclrtFree(deviceId);
  aclFinalize();

  return 0;
}
```
