# aclnnBatchNormGatherStatsWithCounts

## 支持的产品型号
- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型
每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnBatchNormGatherStatsWithCountsGetWorkspaceSize”接口获取入参并根据流程计算所需workspace大小，再调用“aclnnBatchNormGatherStatsWithCounts”接口执行计算。

* `aclnnStatus aclnnBatchNormGatherStatsWithCountsGetWorkspaceSize(const aclTensor* input, const aclTensor* mean, const aclTensor* invstd, aclTensor* runningMean, aclTensor* runningVar, double momentum, double eps, const aclTensor* counts, aclTensor* meanAll, aclTensor* invstdAll, uint64_t* workspaceSize, aclOpExecutor** executor) `
* `aclnnStatus aclnnBatchNormGatherStatsWithCounts(void* workspace, uint64_t workspaceSize, aclOpExecutor* executor, const aclrtStream stream)`
## 功能描述

- 算子功能：
收集所有device的均值和方差，更新全局的均值和方差。BatchNorm的性能和BatchSize相关，BatchSize越大，BatchNorm的统计量也会越准。然而像检测这样的任务，占用显存较高，一张显卡往往只拿较少的图片，比如两张来训练，这就导致BN的表现变差。一个解决方式就是SyncBatchNorm，所有卡共享同一个BatchNorm，得到全局的统计量。
batch_norm_gather_stats_with_counts计算时，依赖batch_norm_stats计算单卡数据的均值和标准差的倒数。

- 计算公式：

$$
    y = \frac{(x-E[x])}{\sqrt{Var(x)+ ε}} * γ + β
$$

runningMean和runningVar更新公式如下：

$$
    runningMean=runningMean*(1-momentum) + E[x]*runningMean
$$

$$
    runningVar=runningVar*(1-momentum) + E[x]*runningVar
$$


## aclnnBatchNormGatherStatsWithCountsGetWorkspaceSize

- **参数说明：**
  * input(const aclTensor \*, 计算输入): Device侧的aclTensor，数据类型仅支持FLOAT，FLOAT16，支持[非连续的Tensor](common/非连续的Tensor.md)，支持的shape和格式有：2维（对应的格式为NC），3维（对应的格式为NCL），4维（对应的格式为NCHW），5维（对应的格式为NCDHW），6-8维（对应的格式为ND，其中第2维固定为channel轴）。
  * mean(const aclTensor \*, 计算输入): 输入数据均值，Device侧的aclTensor，数据类型支持FLOAT，FLOAT16，支持非连续Tensor，[数据格式](common/数据格式.md)为ND，二维Tensor，第一轴对应的大小与input入参中的C轴长度相同。
  * invstd(const aclTensor \*, 计算输入): 输入数据标准差的倒数，Device侧的aclTensor，数据类型支持FLOAT，FLOAT16，支持非连续Tensor，[数据格式](common/数据格式.md)为ND，二维Tensor，一轴对应的大小与input入参中的C轴长度相同，shape与mean对应的shape一致。支持元素值均大于0，小于等于0时，不做保证。
  * runningMean(aclTensor \*, 计算输入): 训练时数据的均值，Device侧的aclTensor，数据类型支持FLOAT，FLOAT16，支持非连续Tensor，[数据格式](common/数据格式.md)为ND，一维Tensor，长度与input入参中的C轴长度相同。
  * runningVar(aclTensor \*, 计算输入): 训练时数据的方差，Device侧的aclTensor，数据类型支持FLOAT，FLOAT16，支持非连续Tensor，[数据格式](common/数据格式.md)为ND，一维Tensor，长度与input入参中的C轴长度相同。支持元素值均大于0，小于等于0时，不做保证。
  * momentum(double \*, 计算输入): runningMean和runningVar的指数平滑参数，默认值为1e-1。
  * eps(double \*, 计算输入): 用于防止产生除0的偏移，host侧的double值，默认值为1e-5。
  * counts(const aclTensor \*, 计算输入): 输入数据元素的个数，Device侧的Tensor，数据类型支持FLOAT，FLOAT16，支持非连续Tensor，[数据格式](common/数据格式.md)为ND，一维Tensor，长度与mean/invstd入参0轴的长度相同。支持元素值均为正整数，其余场景不做保证。
  * meanAll(aclTensor \*, 计算输出): 所有卡上数据的均值，Device侧的aclTensor，数据类型与input一致，支持非连续Tensor，[数据格式](common/数据格式.md)与输入一致。
  * invstdAll(aclTensor \*, 计算输出): 所有卡上数据的标准差的倒数，Device侧的aclTensor，数据类型与input一致，支持非连续Tensor，[数据格式](common/数据格式.md)与输入一致。
  * workspaceSize(uint64_t \*, 出参): 返回用户需要在Device侧申请的workspace大小。
  * executor(aclOpExecutor \*\*, 出参): 返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```
第一段接口完成入参校验，出现以下场景时报错：
161001 (ACLNN_ERR_PARAM_NULLPTR)：1. 传入的指针类型入参是空指针。
161002 (ACLNN_ERR_PARAM_INVALID)：1. input，mean, invstd, runningMeanOptional, runningVarOptional, counts数据类型和数据格式不在支持的范围之内。
                                  2. input，mean, invstd, runningMeanOptional, runningVarOptional, counts数据的shape不在支持的范围内。

```

## aclnnBatchNormGatherStatsWithCounts

- **参数说明：**
  * workspace(void \*, 入参): 在Device侧申请的workspace内存起址。
  * workspaceSize(uint64_t, 入参): 在Device侧申请的workspace大小，由第一段接口aclnnBatchNormGatherStatsWithCounts获取。
  * executor(aclOpExecutor \*, 入参): op执行器，包含了算子计算流程。
  * stream(const aclrtStream , 入参): 指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus: 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
无

## 调用示例
示例代码如下，仅供参考，具体编译和执行过程请参考[编译与运行样例](common/编译与运行样例.md)。
```c++
#include <iostream>
#include <vector>
#include "acl/acl.h"
#include "aclnnop/aclnn_batch_norm_gather_stats_with_counts.h"

#define CHECK_RET(cond, return_expr) \
  do {                               \
    if (!(cond)) {                   \
      return_expr;                   \
    }                                \
  } while (0)

#define LOG_PRINT(message, ...)     \
  do {                              \
    printf(message, ##__VA_ARGS__); \
  } while (0)

int64_t GetShapeSize(const std::vector<int64_t>& shape) {
  int64_t shapeSize = 1;
  for (auto i : shape) {
    shapeSize *= i;
  }
  return shapeSize;
}

int Init(int32_t deviceId, aclrtContext* context, aclrtStream* stream) {
  // 固定写法，AscendCL初始化
  auto ret = aclInit(nullptr);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclInit failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetDevice(deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetDevice failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateContext(context, deviceId);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtSetCurrentContext(*context);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSetCurrentContext failed. ERROR: %d\n", ret); return ret);
  ret = aclrtCreateStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtCreateStream failed. ERROR: %d\n", ret); return ret);
  return 0;
}

template <typename T>
int CreateAclTensor(const std::vector<T>& hostData, const std::vector<int64_t>& shape, void** deviceAddr,
                    aclDataType dataType, aclTensor** tensor) {
  auto size = GetShapeSize(shape) * sizeof(T);
  // 调用aclrtMalloc申请Device侧内存
  auto ret = aclrtMalloc(deviceAddr, size, ACL_MEM_MALLOC_HUGE_FIRST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMalloc failed. ERROR: %d\n", ret); return ret);

  // 调用aclrtMemcpy将Host侧数据拷贝到Device侧内存上
  ret = aclrtMemcpy(*deviceAddr, size, hostData.data(), size, ACL_MEMCPY_HOST_TO_DEVICE);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtMemcpy failed. ERROR: %d\n", ret); return ret);

  // 计算连续tensor的strides
  std::vector<int64_t> strides(shape.size(), 1);
  for (int64_t i = shape.size() - 2; i >= 0; i--) {
    strides[i] = shape[i + 1] * strides[i + 1];
  }

  // 调用aclCreateTensor接口创建aclTensor
  *tensor = aclCreateTensor(shape.data(), shape.size(), dataType, strides.data(), 0, aclFormat::ACL_FORMAT_ND,
                            shape.data(), shape.size(), *deviceAddr);
  return 0;
}

int main() {
  // 1. （固定写法）device/context/stream初始化，参考AscendCL对外接口列表
  // 根据自己的实际device填写deviceId
  int32_t deviceId = 0;
  aclrtContext context;
  aclrtStream stream;
  auto ret = Init(deviceId, &context, &stream);
  // check根据自己的需要处理
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("Init acl failed. ERROR: %d\n", ret); return ret);

  // 2. 构造输入与输出，需要根据API的接口自定义构造
  std::vector<int64_t> inputShape = {2, 4, 2};
  std::vector<int64_t> meanShape = {4, 4};
  std::vector<int64_t> rMeanShape = {4};
  std::vector<int64_t> rVarShape = {4};
  std::vector<int64_t> countsShape = {4};
  std::vector<int64_t> invstdShape = {4, 4};
  std::vector<int64_t> meanAllShape = {4};
  std::vector<int64_t> invstdAllShape = {4};
  void* inputDeviceAddr = nullptr;
  void* meanDeviceAddr = nullptr;
  void* rMeanDeviceAddr = nullptr;
  void* rVarDeviceAddr = nullptr;
  void* countsDeviceAddr = nullptr;
  void* invstdDeviceAddr = nullptr;
  void* meanAllDeviceAddr = nullptr;
  void* invstdAllDeviceAddr = nullptr;
  aclTensor* input = nullptr;
  aclTensor* mean = nullptr;
  aclTensor* rMean = nullptr;
  aclTensor* rVar = nullptr;
  aclTensor* counts = nullptr;
  aclTensor* invstd = nullptr;
  aclTensor* meanAll = nullptr;
  aclTensor* invstdAll = nullptr;
  std::vector<float> inputHostData = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
  std::vector<float> meanHostData = {0, 1, 2, 3, 4, 5, 6, 7};
  std::vector<float> rMeanHostData = {1, 2, 3, 4};
  std::vector<float> rVarHostData = {5, 6, 7, 8};
  std::vector<float> countsHostData = {1, 2, 3, 4};
  std::vector<float> invstdHostData = {1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4};
  std::vector<float> meanAllHostData = (4, 0);
  std::vector<float> invstdAllHostData = (4, 0);
  double momentum = 1e-2;
  double eps = 1e-4;

  // 创建input aclTensor
  ret = CreateAclTensor(inputHostData, inputShape, &inputDeviceAddr, aclDataType::ACL_FLOAT, &input);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建mean aclTensor
  ret = CreateAclTensor(meanHostData, meanShape, &meanDeviceAddr, aclDataType::ACL_FLOAT, &mean);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建rMean aclTensor
  ret = CreateAclTensor(rMeanHostData, rMeanShape, &rMeanDeviceAddr, aclDataType::ACL_FLOAT, &rMean);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建rVar aclTensor
  ret = CreateAclTensor(rVarHostData, rVarShape, &rVarDeviceAddr, aclDataType::ACL_FLOAT, &rVar);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建counts aclTensor
  ret = CreateAclTensor(countsHostData, countsShape, &countsDeviceAddr, aclDataType::ACL_FLOAT, &counts);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建invstd aclTensor
  ret = CreateAclTensor(invstdHostData, invstdShape, &invstdDeviceAddr, aclDataType::ACL_FLOAT, &invstd);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建meanAll aclTensor
  ret = CreateAclTensor(meanAllHostData, meanAllShape, &meanAllDeviceAddr, aclDataType::ACL_FLOAT, &meanAll);
  CHECK_RET(ret == ACL_SUCCESS, return ret);
  // 创建invstdAll aclTensor
  ret = CreateAclTensor(invstdAllHostData, invstdAllShape, &invstdAllDeviceAddr, aclDataType::ACL_FLOAT, &invstdAll);
  CHECK_RET(ret == ACL_SUCCESS, return ret);

  uint64_t workspaceSize = 0;
  aclOpExecutor* executor;

  // aclnnBatchNormGatherStatsWithCounts接口调用示例
  // 3. 调用CANN算子库API，需要修改为具体的API名称
  // 调用aclnnBatchNormGatherStatsWithCounts第一段接口
  ret = aclnnBatchNormGatherStatsWithCountsGetWorkspaceSize(input, mean, invstd, rMean, rVar, momentum, eps, counts, meanAll, invstdAll, &workspaceSize, &executor);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnBatchNormGatherStatsWithCountsGetWorkspaceSize failed. ERROR: %d\n", ret); return ret);
  // 根据第一段接口计算出的workspaceSize申请device内存
  void* workspaceAddr = nullptr;
  if (workspaceSize > 0) {
    ret = aclrtMalloc(&workspaceAddr, workspaceSize, ACL_MEM_MALLOC_HUGE_FIRST);
    CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("allocate workspace failed. ERROR: %d\n", ret); return ret);
  }
  // 调用aclnnBatchNormGatherStatsWithCounts第二段接口
  ret = aclnnBatchNormGatherStatsWithCounts(workspaceAddr, workspaceSize, executor, stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclnnBatchNormGatherStatsWithCounts failed. ERROR: %d\n", ret); return ret);

  // 4. （固定写法）同步等待任务执行结束
  ret = aclrtSynchronizeStream(stream);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("aclrtSynchronizeStream failed. ERROR: %d\n", ret); return ret);

  // 5. 获取输出的值，将Device侧内存上的结果拷贝至Host侧，需要根据具体API的接口定义修改
  auto size = GetShapeSize(meanAllShape);
  std::vector<float> resultData(size, 0);
  ret = aclrtMemcpy(resultData.data(), resultData.size() * sizeof(resultData[0]), meanAllDeviceAddr,
                    size * sizeof(resultData[0]), ACL_MEMCPY_DEVICE_TO_HOST);
  CHECK_RET(ret == ACL_SUCCESS, LOG_PRINT("copy result from device to host failed. ERROR: %d\n", ret); return ret);
  for (int64_t i = 0; i < size; i++) {
    LOG_PRINT("result[%ld] is: %f\n", i, resultData[i]);
  }

  // 6. 释放aclTensor和aclScalar，需要根据具体API的接口定义修改
  aclDestroyTensor(input);
  aclDestroyTensor(mean);
  aclDestroyTensor(rMean);
  aclDestroyTensor(rVar);
  aclDestroyTensor(counts);
  aclDestroyTensor(invstd);
  aclDestroyTensor(meanAll);
  aclDestroyTensor(invstdAll);

  // 7. 释放Device资源，需要根据具体API的接口定义修改
  aclrtFree(inputDeviceAddr);
  aclrtFree(meanDeviceAddr);
  aclrtFree(rMeanDeviceAddr);
  aclrtFree(rVarDeviceAddr);
  aclrtFree(countsDeviceAddr);
  aclrtFree(invstdDeviceAddr);
  aclrtFree(meanAllDeviceAddr);
  aclrtFree(invstdAllDeviceAddr);
  if (workspaceSize > 0) {
    aclrtFree(workspaceAddr);
  }
  aclrtDestroyStream(stream);
  aclrtDestroyContext(context);
  aclrtResetDevice(deviceId);
  aclFinalize();
  return 0;
}
```
