# aclnnClampMin

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnClampMinGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnClampMin”接口执行计算。

- `aclnnStatus aclnnClampMinGetWorkspaceSize(const aclTensor *self, const aclScalar* min, aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **executor)`
- `aclnnStatus aclnnClampMin(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

- 算子功能：将输入的所有元素限制在[min, inf]范围内。

- 计算公式：

  $$
  {y}_{i} = max({{x}_{i}},{min\_value}_{i})
  $$

## aclnnClampMinGetWorkspaceSize
- **参数说明：**

  - self(const aclTensor *): 输入tensor，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持）。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - min(const aclScalar*): 下界，数据类型需要可转换成self的数据类型。
  - out(aclTensor *)：输出，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持），且数据类型和self保持一致，shape和self保持一致，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t *)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor **)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的self、out其中一个为空指针，min为空指针。
  ACLNN_ERR_PARAM_INVALID: 1. self、out的数据类型和数据格式不在支持的范围之内。
                           2. self的数据类型与输出out的类型不一致。
  ```

## aclnnClampMin

- **参数说明：**

  - workspace(void *)：在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t)：在Device侧申请的workspace大小，由第一段接口aclnnClampMinGetWorkspaceSize获取。
  - executor(aclOpExecutor)：op执行器，包含了算子计算流程。
  - stream(const aclrtStream)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制
无
