# aclnnClampMinTensor&aclnnInplaceClampMinTensor

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

- aclnnClampMinTensor和aclnnInplaceClampMinTensor实现相同的功能，使用区别如下，请根据自身实际场景选择合适的算子。

  - aclnnClampMinTensor：需新建一个输出张量对象存储计算结果。
  - aclnnInplaceClampMinTensor：无需新建输出张量对象，直接在输入张量的内存中存储计算结果。
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnClampMinTensorGetWorkspaceSize”或者“aclnnInplaceClampMinTensorGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnClampMinTensor”或者“aclnnInplaceClampMinTensor”接口执行计算。

  - `aclnnStatus aclnnClampMinTensorGetWorkspaceSize(const aclTensor *self, const aclTensor* min, aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **executor)`
  - `aclnnStatus aclnnClampMinTensor(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`
  - `aclnnStatus aclnnInplaceClampMinTensorGetWorkspaceSize(const aclTensor *selfRef, const aclTensor* min, uint64_t *workspaceSize, aclOpExecutor **executor)`
  - `aclnnStatus aclnnInplaceClampMinTensor(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

- 算子功能：将输入的所有元素限制在[min, inf]范围内。

- 计算公式：

  $$
  {y}_{i} = max({{x}_{i}},{min\_value}_{i})
  $$

## aclnnClampMinTensorGetWorkspaceSize
- **参数说明**：

  - self(const aclTensor *): 输入tensor，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（Atlas A2系列产品/Atlas 300I A2推理产品）。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - min(const aclTensor*): 输入下限值tensor，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（Atlas A2系列产品/Atlas 300I A2推理产品），且数据类型需要与self、max满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)）。shape需要与self、max满足[broadcast关系](common/broadcast关系.md)，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - out(aclTensor *)：输出，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（Atlas A2系列产品/Atlas 300I A2推理产品），且数据类型和self保持一致，shape和self保持一致，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t *)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor **)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的self、out其中一个为空指针，min为空指针。
  ACLNN_ERR_PARAM_INVALID: 1. self、out的数据类型和数据格式不在支持的范围之内。
                           2. self的数据类型与输出out的类型不一致。
  ```

## aclnnClampMinTensor

- **参数说明：**

  - workspace(void *)：在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t)：在Device侧申请的workspace大小，由第一段接口aclnnClampMinTensorGetWorkspaceSize获取。
  - executor(aclOpExecutor)：op执行器，包含了算子计算流程。
  - stream(const aclrtStream)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceClampMinTensorGetWorkspaceSize

- **参数说明：**

  - selfRef(const aclTensor *): 输入输出tensor，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（Atlas A2系列产品/Atlas 300I A2推理产品）。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - min(const aclTensor*): 输入下限值tensor，数据类型支持FLOAT16、FLOAT、FLOAT64、INT8、UINT8、INT16、INT32、INT64、BFLOAT16（Atlas A2系列产品/Atlas 300I A2推理产品），且数据类型需要与self、max满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)）。shape需要与self、max满足[broadcast关系](common/broadcast关系.md)，支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，维度不超过8维。
  - workspaceSize(uint64_t *)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor **)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的selfRef为空指针，min为空指针。
  ACLNN_ERR_PARAM_INVALID: 1. selfRef的数据类型和数据格式不在支持的范围之内。
                           2. selfRef和min的shape不满足broadcast关系。
                           3. selfRef和min类型推导失败。
                           4. selfRef和min的维度大小超过8。
  ```

## aclnnInplaceClampMinTensor

- **参数说明：**

  - workspace(void *)：在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t)：在Device侧申请的workspace大小，由第一段接口aclnnInplaceClampMinTensorGetWorkspaceSize获取。
  - executor(aclOpExecutor)：op执行器，包含了算子计算流程。
  - stream(const aclrtStream)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无