# aclnnHardswish&aclnnInplaceHardswish

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型
+ aclnnHardswish和aclnnInplaceHardswish实现相同的功能，使用区别如下，请根据自身实际场景选择合适的算子。
  + aclnnHardswish：需新建一个输出张量对象存储计算结果。
  + aclnnInplaceHardswish：无需新建输出张量对象，直接在输入张量的内存中存储计算结果。

+ 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnHardswishGetWorkspaceSize”或者“aclnnInplaceHardswishGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnHardswish”或者"aclnnInplaceHardswish"接口执行计算。

  + `aclnnStatus aclnnHardswishGetWorkspaceSize(const aclTensor *self, const aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **opExecutor)`
  + `aclnnStatus aclnnHardswish(void *workspace, uint64_t workspaceSize, const aclrtStream stream, aclOpExecutor *opExecutor)`
  + `aclnnStatus aclnnInplaceHardswishGetWorkspaceSize(const aclTensor *self, uint64_t *workspaceSize, aclOpExecutor **opExecutor)`
  + `aclnnStatus aclnnInplaceHardswish(void *workspace, uint64_t workspaceSize, const aclrtStream stream, aclOpExecutor *opExecutor)`

## 功能描述

+ 算子功能：
  激活函数， 返回与输入tensor shape相同的输出tensor，输入的value小于-3时取0，大于3时取该value，其余时刻取value加3的和乘上value再除以6。
+ 计算公式：
  
  $$
  Hardswish(x)=\begin{cases}
  x, & x\gt3 \\
  0, &  x\le -3 \\
  \frac{x · (x + 3)}{6}, & otherwise
  
   \end{cases}
  $$

## aclnnHardswishGetWorkspaceSize

+ **参数说明：**

  + self(const aclTensor *self, 计算输入): Device侧的aclTensor，数据类型支持 FLOAT、FLOAT16，支持非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
  + out(const aclTensor *self, 计算输出): Device侧的aclTensor，数据类型支持 FLOAT、FLOAT16，支持非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
  + workspaceSize(uint64_t\*, 出参): 返回用户需要在Device侧申请的workspace大小。
  + executor(aclOpExecutor\**, 出参): 返回op执行器，包含了算子计算流程计算流程。

+ **返回值：**

  aclnnStatus： 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnHardswish

+ **参数说明：**

  + workspace(void\*, 入参)：在Device侧申请的workspace内存起址。
  + workspaceSize(uint64_t, 入参): 在Device侧申请的workspace大小，由第一段接口aclnnHardswishGetWorkspaceSize获取。
  + stream(aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。
  + executor(aclOpExecutor\*, 入参)：op执行器，包含了算子计算流程。

+ **返回值：**

  aclnnStatus： 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceHardswishGetWorkspaceSize

+ **参数说明：**

  + self (const aclTensor *, 计算输入|计算输出): Device侧的aclTensor，数据类型支持 FLOAT、FLOAT16，支持非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
  + workspaceSize (uint64_t\*, 出参): 返回用户需要在Device侧申请的workspace大小。
  + executor (aclOpExecutor\**, 出参): 返回op执行器，包含了算子计算流程计算流程。

+ **返回值：**

  aclnnStatus： 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceHardswish

+ **参数说明：**

  + workspace(void\*, 入参)：在Device侧申请的workspace内存起址。
  + workspaceSize(uint64_t, 入参): 在Device侧申请的workspace大小，由第一段接口aclnnInplaceHardswishGetWorkspaceSize获取。
  + stream(aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。
  + executor(aclOpExecutor\*, 入参)：op执行器，包含了算子计算流程。

+ **返回值：**
  aclnnStatus： 返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无