# aclnnInplaceBitwiseAndTensor

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnInplaceBitwiseAndTensorGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnInplaceBitwiseAndTensor”接口执行计算。

- `aclnnStatus aclnnInplaceBitwiseAndTensorGetWorkspaceSize(const aclTensor *selfRef, const aclTensor *other, uint64_t *workspaceSize, aclOpExecutor **executor)`
- `aclnnStatus aclnnInplaceBitwiseAndTensor(void *workspace, uint64_t workspaceSize, const aclStream stream, aclOpExecutor **executor)`

## 功能描述

- 算子功能：输入为BOOL型tensor时，进行逻辑与运算；输入为INT型时，进行位与运算。
- 计算公式：

  $$
  selfRef_i = selfRef_i \& other
  $$

## aclnnInplaceBitwiseAndTensorGetWorkspaceSize

- **参数说明：**

  - selfRef(aclTensor*, 计算输入)：公式中的```selfRef```，数据类型支持INT16,INT32,INT64,INT8,UINT8,BOOL，且数据类型需要与other的数据类型满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与other满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - other(aclTensor*, 计算输入)：公式中的```other```，数据类型支持INT16,INT32,INT64,INT8,UINT8,BOOL，且数据类型需要与self构成满足数据类型推导规则（参见[互推导关系](common/互推导关系.md)），shape需要与self满足[broadcast关系](common/broadcast关系.md)。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t*, 出参)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor*, 出参)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  161001(ACLNN_ERR_PARAM_NULLPTR)：1. 传入的selfRef、other或out是空指针。
  161002(ACLNN_ERR_PARAM_INVALID)：1. selfRef和other的数据类型不在支持的范围之内。
                                   2. selfRef和other不满足数据类型推导规则。
                                   3. selfRef和other推导出的数据类型无法转换为指定输出selfRef的类型。
                                   4. selfRef和other的shape无法做broadcast。
  ```

## aclnnInplaceBitwiseAndTensor

- **参数说明：**

  - workspace(void*, 入参)：在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t, 入参)：在Device侧申请的workspace大小，由第一段接口aclnnInplaceBitwiseAndTensorGetWorkspaceSize获取。
  - executor(aclOpExecutor*, 入参)：op执行器，包含了算子计算流程。
  - stream(aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无
