# aclnnKthvalue

## 接口原型
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnXxxGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnXxx”接口执行计算。

* `aclnnStatus aclnnKthvalueGetWorkspaceSize(const aclTensor *self, int64_t k, int64_t dim, bool keepdim, aclTensor *valuesOut, aclTensor *indicesOut, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclnnKthvalue(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

算子功能: 返回输入Tensor在指定维度上的第k个最小值及索引。

### 接口参数

**接口：** **aclnnKthvalueGetWorkspaceSize**
输入：

* self： Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、INT32。支持连续和非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
* k：int64_t类型整数。表示取指定维度上第k个最小值。取值范围为[0, self.size(dim)]。
* dim ：int64_t类型整数。表示取输入张量的指定维度，如果没有给出，则选择最后一个dim。取值范围为[-self.dim(), self.dim()]。
* keepdim：bool类型数据，表示输出张量是否保留了dim。True表示valuesOut和indicesOut张量的大小都与self相同，False表示dim将被压缩，得到的张量维数比input少1。
* valuesOut：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、INT32，且数据类型与self保持一致。支持连续和非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
* indicesOut：Device侧的aclTensor，数据类型支持INT64。表示原始输入张量中沿dim维的第k个最小值的下标。支持连续和非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
* workspaceSize：返回用户需要在Device侧申请的workspace大小。
* executor：返回op执行器，包含了算子计算流程。

输出：

* aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)

```
ACLNN_ERR_PARAM_NULLPTR：
1. 传入的self、valuesOut或indicesOut是空指针。

ACLNN_ERR_PARAM_INVALID：
1. self、valuesOut或indicesOut的数据类型和数据格式不在支持的范围之内。
2. dim不在输入self的合理维度范围内。
3. k小于0或者k大于输入self在dim维度上的size大小。
```

**接口：** **aclnnKthvalue**
输入：

* workspace：在Device侧申请的workspace内存起址。
* workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnKthvalueGetWorkspaceSize获取。
* executor：op执行器，包含了算子计算流程。
* stream：指定执行任务的 AscendCL Stream流。

输出：

* aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)
