# aclnnOneHot

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnOneHotGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnOneHot”接口执行计算。

- `aclnnStatus aclnnOneHotGetWorkspaceSize(const aclTensor* self, int numClasses, const aclTensor* onValue, const aclTensor* offValue, int64_t axis, aclTensor* out, uint64_t* workspaceSize, aclOpExecutor** executor)`

- `aclnnStatus aclnnOneHot(void* workspace, uint64_t workspaceSize, aclOpExecutor* executor, aclrtStream stream)`

## 功能描述

- 算子功能：对长度为n的输入self， 经过one_hot的计算后得到一个大小为[n*k]的输出out，其中k的值为numClasses。
- 输出的元素满足下列公式：
  $$
  out[i][j]=\left\{
  \begin{aligned}
  onValue,\quad self[i] = j \\
  offValue, \quad self[i] \neq j
  \end{aligned}
  \right.
  $$

## aclnnOneHotGetWorkspaceSize

- **参数说明：**

  - self(const aclTensor*, 计算输入)：输入的self， 数据类型支持INT32、INT64，支持非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
- numClasses(int， 计算输入)：输入的numClasses，数据类型必须输入INT64。最小值为0。若numClasses的最小值为0，则返回空Tensor。如果self存在元素大于numClasses，这些元素会被编码成全0的tensor。
  - onValue(const aclTensor*，计算输入)：公式中的onValue，数据类型支持INT64。
- offValue(const aclTensor*，计算输入)：公式中的offValue，数据类型支持INT64。
  - axis(int64_t，计算输入)：最小值为-1，最大值为self的维度数。若值为-1，编码向量会往self的最后一维插入。
- out(aclTenosr*, 计算输出)：公式中的输出out，数据类型支持INT32、INT64。支持非连续的Tensor，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t*, 出参)：返回用户需要在Device侧申请的workspace大小。
- executor(aclOpExecutor**, 出参)：返回op执行器，包含了算子计算流程。


- **返回码：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

  ```
  ACLNN_ERR_PARAM_NULLPTR： 1. 传入的selfRef为空指针。
  ACLNN_ERR_PARAM_INVALID： 1. self为空，且numClasses小于等于0。
                            2. numClasses小于-1。
  ```

## aclnnOneHot

- **参数说明：**

  * workspace(void* 入参)：在Device侧申请的workspace内存起址。

  * workspaceSize(uint64_t 入参)：在Device侧申请的workspace大小，由第一段接口aclnnInplaceOneGetWorkspaceSize获取。

  * executor(aclOpExecutor* 入参)：op执行器，包含了算子计算流程。

  * stream(aclrtStream 入参)：指定执行任务的 AscendCL Stream流。


- **返回码：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无。
