# aclnnSilu

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnSiluGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnSilu”接口执行计算。

* `aclnnStatus aclnnSiluGetWorkspaceSize(const aclTensor *self, aclTensor *out, uint64_t *workspaceSize, aclOpExecutor **executor)`
* `aclnnStatus aclnnSilu(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclStream stream)`

## 功能描述
- 算子功能：该算子也被称为Swish函数（$\beta = 1$），对输入tensor逐元素进行Silu函数运算并输出结果tensor。

- 计算公式：

$$
output_i = \frac {input_i} {1 + e^{-input_i}}
$$



## aclnnSiluGetWorkspaceSize

- **参数说明：**

  * self(const aclTensor)：Device侧的tensor，数据类型支持FLOAT16、FLOAT、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持）。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。

  * out(aclTensor)：Device侧的tensor，数据类型支持FLOAT16、FLOAT、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持）。支持[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND，且类型与shape需要与self一致。

  * workspaceSize(int64_t \*)：返回用户需要在Device侧申请的workspace大小。

  * executor(aclOpExecuter \**)：返回op执行器，包含了算子计算流程。


- **返回码：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

```c++
161001 ACLNN_ERR_PARAM_NULLPTR: 1.传入的self或out是空指针。
161002 ACLNN_ERR_PARAM_INVALID: 1.self和out的数据类型和数据类型不在支持的范围内。
                                2.self和ou的shape不匹配。
                                3.self和out的type不匹配。
```

## aclnnSilu

- **参数说明：**

  * workspace(void \*)：在Device侧申请的workspace内存地址。

  * workspaceSize(uint64_t)：在Device侧申请的workspace大小，由第一段接口aclnnSiluGetWorkspaceSize获取。

  * executer(aclOpExecutor)：op执行器，包含了算子计算流程。

  * stream：指定执行任务的 AscendCL Stream流。


- **返回码：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无。
