# aclnnTanh&aclnnInplaceTanh

## 支持的产品型号

- Atlas 训练产品。
- Atlas A2系列产品/Atlas 300I A2推理产品。

## 接口原型

- aclnnTanh和aclnnInplaceTanh实现相同的功能，使用区别如下，请根据自身实际场景选择合适的算子。

  - aclnnTanh：需新建一个输出张量对象存储计算结果。
  - aclnnInplaceTanh：无需新建输出张量对象，直接在输入张量的内存中存储计算结果。 

- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnTanhGetWorkspaceSize”或者“aclnnInplaceTanhGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnTanh”或者“aclnnInplaceTanh”接口执行计算。

  + `aclnnStatus aclnnTanhGetWorkspaceSize(const aclTensor *self, aclTensor *out, uint64_t &workspaceSize, aclOpExecutor **opExecutor)`
  + `aclnnStatus aclnnTanh(void *workspace, uint64_t workspaceSize, const aclrtStream stream, aclOpExecutor *opExecutor)`
  + `aclnnStatus aclnnInplaceTanhGetWorkspaceSize(const aclTensor *self, uint64_t *workspaceSize, aclOpExecutor **executor)`
  + `aclnnStatus aclnnInplaceTanh(void *workspace, uint64_t workspaceSize, const aclrtStream stream, aclOpExecutor *opExecutor)`

## 功能描述
+ 算子功能：
  激活函数， 返回与输入tensor shape相同的tensor，对输入tensor进行elmentwise的计算。
+ 计算公式：

  $$
  tanh(x)=\frac{e^{x}-e^{-x}}{e^x + e^{-x}}
  $$

## aclnnTanhGetWorkspaceSize

- **参数说明：**

  - self(const aclTensor\*, 计算输入)：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BOOL、UINT8、INT8、INT16、INT32、INT64、BLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品)。支持非连续tensor，[数据格式](common/数据格式.md)支持ND。
  - out(aclTensor\*, 计算输出)：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BOOL、UINT8、INT8、INT16、INT32、INT64、BLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)。支持非连续tensor，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t\*, 出参)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor\*\*, 出参)：返回op执行器，包含了算子计算流程。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。
  ```
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的self、out是空指针。
  ACLNN_ERR_PARAM_INVALID: 1. self和out的数据类型和数据格式不在支持的范围之内。
  ```

## aclnnTanh

- **参数说明：**

  - workspace(void \*, 入参): 在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t, 入参)：在Device侧申请的workspace大小，由第一段接口aclnnTanhGetWorkspaceSize获取。
  - executor(aclOpExecutor \*, 入参)：op执行器，包含了算子计算流程。
  - stream(const aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## aclnnInplaceTanhGetWorkspaceSize

- **参数说明：**

  - self(const aclTensor\*, 计算输入|计算输出)：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、BOOL、UINT8、INT8、INT16、INT32、INT64、BLOAT16(仅Atlas A2系列产品/Atlas 300I A2推理产品支持)。支持非连续tensor，[数据格式](common/数据格式.md)支持ND。
  - workspaceSize(uint64_t\*, 出参)：返回用户需要在Device侧申请的workspace大小。
  - executor(aclOpExecutor\*\*, 出参)：返回op执行器，包含了算子计算流程。
  
- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。
  ```
  ACLNN_ERR_PARAM_NULLPTR: 1. 传入的self、out是空指针。
  ACLNN_ERR_PARAM_INVALID: 1. self和out的数据类型和数据格式不在支持的范围之内。
  ```

## aclnnInplaceTanh

- **参数说明：**

  - workspace(void \*, 入参): 在Device侧申请的workspace内存起址。
  - workspaceSize(uint64_t, 入参)：在Device侧申请的workspace大小，由第一段接口aclnnInplaceTanhGetWorkspaceSize获取。
  - executor(aclOpExecutor \*, 入参)：op执行器，包含了算子计算流程。
  - stream(const aclrtStream, 入参)：指定执行任务的 AscendCL Stream流。

- **返回值：**

  aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)。

## 约束与限制

无