### aclnnTopk

## 接口原型
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnXxxGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnXxx”接口执行计算。

- `aclnnStatus aclnnTopkGetWorkspaceSize(const aclTensor *self, int64_t k, int64_t dim, bool largest, bool sorted, aclTensor *valuesOut, aclTensor *indicesOut, uint64_t *workspaceSize, aclOpExecutor **executor)`
- `aclnnStatus aclnnTopk(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

算子功能: 返回输入Tensor在指定维度上的k个极值及索引。

#### 接口参数

**接口：** `aclnnTopkGetWorkspaceSize`

输入：

- self： Device侧的aclTensor，数据类型支持INT8、UINT8、INT16、INT32、INT64、FLOAT16、FLOAT32、DOUBLE、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持）。支持连续和[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
- k：int64_t类型整数。表示计算维度上输出的极值个数。取值范围为[0, self.size(dim)]。
- dim ：int64_t类型整数。表示计算维度。取值范围为[-self.dim(), self.dim())。
- largest：bool类型数据。True表示计算维度上的结果应由大到小输出，False表示计算维度上的结果由小到大输出。
- sorted：bool类型数据。True表示输出结果需要排序（若largest为True则结果从大到小排序，若largest为Flase则结果从小到大排序），False表示输出结果不排序，按输入时的数据顺序输出。
- valuesOut：Device侧的aclTensor，数据类型支持INT8、UINT8、INT16、INT32、INT64、FLOAT16、FLOAT32、DOUBLE、BFLOAT16（仅Atlas A2系列产品/Atlas 300I A2推理产品支持），且数据类型与self保持一致。支持连续和[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
- indicesOut：Device侧的aclTensor，数据类型支持INT64。支持连续和[非连续的Tensor](common/非连续的Tensor.md)，[数据格式](common/数据格式.md)支持ND。
- workspaceSize：返回用户需要在Device侧申请的workspace大小。
- executor：返回op执行器，包含了算子计算流程。

输出：

- aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)

```
ACLNN_ERR_PARAM_NULLPTR：
1. 传入的self、valuesOut或indicesOut是空指针。

ACLNN_ERR_PARAM_INVALID：
1. self、valuesOut或indicesOut的数据类型和数据格式不在支持的范围之内。
2. dim不在输入self的合理维度范围内。
3. k小于0或者k大于输入self在dim维度上的size大小。
```

**接口：** `aclnnTopk`

输入：

- workspace：在Device侧申请的workspace内存起址。
- workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnTopkGetWorkspaceSize获取。
- executor：op执行器，包含了算子计算流程。
- stream：指定执行任务的 AscendCL Stream流。

输出：

- aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)
