### aclnnUpsampleBilinear2d

## 接口原型
- 每个算子分为[两段式接口](common/两段式接口.md)，必须先调用“aclnnXxxGetWorkspaceSize”接口获取计算所需workspace大小以及包含了算子计算流程的执行器，再调用“aclnnXxx”接口执行计算。

- `aclnnStatus aclnnUpsampleBilinear2dGetWorkspaceSize( const aclTensor *self, const aclIntArray *outputSize, bool alignCorners, double scalesH, double scalesW, aclTensor *outRef, uint64_t *workspaceSize, aclOpExecutor **executor)`
- `aclnnStatus aclnnUpsampleBilinear2d(void *workspace, uint64_t workspaceSize, aclOpExecutor *executor, const aclrtStream stream)`

## 功能描述

算子功能：对由多个输入通道组成的输入信号应用 2D 双线性上采样。

对于输入shape有三种情况：
（1）指定了scalesH, scalesW，如果输入shape为（N, C, H，W)，则输出shape为（N，C，$\lfloor H * scalesH\rfloor$，$\lfloor W * scalesW\rfloor$）
（2）指定了输出outputSize，如果输入shape为（N, C, H，W)，则输出shape为（N，C，outputSizep[0]，outputSizep[1]）
（3）同时制定了outputSize 和scalesH,scalesW，以outputSize为准。

对于中心对齐的选择：
当align_corners = True时，像素被视为网格的格子上的点,拐角处的像素对齐.可知是点之间是等间距的
当align_corners = False时, 像素被视为网格的交叉线上的点, 拐角处的点依然是原图像的拐角像素,但是差值的点间却按照上图的取法取,导致点与点之间是不等距的，下图为上采样两倍的样例：

$$
\begin{matrix}
   1 & 2 & 3 \\
   4 & 5 & 6 \\
   7 & 8 & 9
  \end{matrix} \tag{原图}
$$

$$
\begin{matrix}
   1 & 1.25 & 1.75& 2.25& 2.75 & 3 \\
   1.75 & 2 & 2.5 & 3 & 3.5 & 3.75 \\
   3.25 & 3.5 & 4 &4.5 & 5 & 5.25 \\
   4.75 & 5 & 5.5 & 6 & 6.5 & 6.75 \\
   6.25 & 6.5 &7 & 7.5 & 8 & 8.25 \\
   7 & 7.25 & 7.75 & 8.25 & 8.75 & 9
  \end{matrix}  \tag{align\_corners = False}
$$

$$
\begin{matrix}
   1 & 1.5 & 3 & 2.5 & 3 \\
   2.5 & 3 & 3.5 & 4 & 4.5 \\
   4 & 5.5 & 5 & 5.5 & 6 \\
   5.5 & 6 & 6.5 & 7 & 7.5 \\
   7 & 7.5 & 8 & 8.5 & 9
  \end{matrix}  \tag{align\_corners = True}
$$

#### 接口参数

**接口：**`aclnnUpsampleBilinear2dGetWorkspaceSize`
输入：

- self：Device侧的aclTensor，数据类型支持FLOAT、FLOAT16、DOUBLE，支持非连续Tensor，[数据格式](common/数据格式.md)支持NCHW, NHWC。
- outputSize：Device侧的aclIntArrat，输出空间大小。
- alignCorners：bool常量，如果设置为`True`，则输入和输出张量按其角像素的中心点对齐，保留角像素处的值。如果设置为`False`，则输入和输出张量通过其角像素的角点对齐，并且插值使用边缘值填充用于边界外值，使此操作在保持不变时独立于输入大小scale_w,scale_h。
- scalesH：double常量，空间大小的height维度乘数。
- scalesW:  double常量，空间大小的width维度乘数。
- outRef: Device侧的aclTensor,  数据类型支持FLOAT、FLOAT16、DOUBLE，支持非连续Tensor，[数据格式](common/数据格式.md)支持NCHW, NHWC。
- workspaceSize：返回用户需要在Device侧申请的workspace大小。
- executor：返回op执行器，包含了算子计算流程。

输出：

- aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)

```
ACLNN_ERR_PARAM_NULLPTR：1. 传入的 self 是空指针时。
ACLNN_ERR_PARAM_INVALID：1. self 的数据类型不在支持的范围之内。
                         2. self 的数据格式不在支持的范围之内。
```

**接口：**`aclnnUpsampleBilinear2d`
输入：

- workspace：在Device侧申请的workspace内存起址。
- workspaceSize：在Device侧申请的workspace大小，由第一段接口aclnnUpsampleBilinear2DGetWorkspaceSize获取。
- stream：指定执行任务的 AscendCL Stream流。
- executor：op执行器，包含了算子计算流程。
  输出：
- aclnnStatus：返回状态码，具体参见[aclnn返回码](common/aclnn返回码.md)
