# 数据格式

数据格式（format）是用于描述一个多维Tensor的轴的业务语义，表示数据的物理排布格式，定义了解读数据的维度，比如1D、2D、3D、4D、5D等。一般在CNN（Convolutional Neural Network）类的API中需要给出具体的数据排布格式。

>**说明：** 
>-   数据格式一般形式为“ACL\_FORMAT\__XXXX_”，为方便描述，可简写为“_XXXX_”。
>-   数据格式中维度含义：N（Batch）表示批量大小、H（Height）表示特征图高度、W（Width）表示特征图宽度、C（Channels）表示特征图通道、D（Depth）表示特征图深度。

## 常见数据格式

目前大部分算子API都是支持ND数据格式的。例如aclnnAdd接口，均标明支持的数据格式是ND（即多维Tensor的，低维度优先连续排布的规则）。

对于aclnnConvolution，该API要求输入的aclTensor设置具有业务语义的格式，而不是ND格式。此类算子，在计算过程中需要知道Tensor中的业务语义才可以进行对应的计算。例如2D卷积中，需要知道Batch维度、Channel维度、Height维度、Width维度与Tensor维度的对应关系是什么。

当调用aclCreateTensor接口的时候，根据API要求设置Tensor的数据格式，当前支持的数据格式有：

ACL\_FORMAT\_ND、ACL\_FORMAT\_NCHW、ACL\_FORMAT\_NHWC、ACL\_FORMAT\_HWCN、ACL\_FORMAT\_NDHWC、ACL\_FORMAT\_NCDHW、ACL\_FORMAT\_NC、ACL\_FORMAT\_NCL。

对于非ND的Tensor，Tensor的维度要求与format表述是一致的。例如：

-   5D的Tensor：要求为ACL\_FORMAT\_NCDHW、ACL\_FORMAT\_NDHWC或者ACL\_FORMAT\_ND（如果API参数说明中未标明支持ND，设置ND格式后API将会校验报错）。
-   4D的Tensor：要求为ACL\_FORMAT\_NCHW、ACL\_FORMAT\_NHWC、ACL\_FORMAT\_HWCN或者ACL\_FORMAT\_ND。
-   3D的Tensor：要求为ACL\_FORMAT\_NCL或者ACL\_FORMAT\_ND。
-   2D的Tensor：要求为ACL\_FORMAT\_NC或者ACL\_FORMAT\_ND。
-   其他维度的Tensor：要求为ACL\_FORMAT\_ND。

## 私有数据格式

除了常见的数据格式，还存在其他的数据格式，比如ACL\_FORMAT\_NC1HWC0、ACL\_FORMAT\_FRACTAL\_Z、ACL\_FORMAT\_NC1HWC0\_C04、ACL\_FORMAT\_FRACTAL\_NZ、ACL\_FORMAT\_NDC1HWC0、ACL\_FORMAT\_FRACTAL\_Z\_3D等。

这些格式，属于NPU的一些私有格式，当前绝大多数的aclnn API还不支持这些格式。如果有个别API申明了支持的数据格式，请以该API实际描述为准。